#include "config.h"
#include "adoc_sem.h"
#undef DEBUG


#ifdef DEBUG
#include <stdio.h>
#include <stdlib.h>
#endif


#ifdef USE_IPC

int  adoc_sem_init(int  *id,  int  pshared,   unsigned   int val){
#ifdef SEMUN_NOT_OK
    union semun {
		int val;
		struct semid_ds *buf;
		ushort * array;
	} argument;
#else
    union semun argument;
#endif
    argument.val = val;


  *id=semget(IPC_PRIVATE, 1, IPC_CREAT|IPC_EXCL|0600);
  if(*id<0)
    return *id;
  return semctl(*id, 0, SETVAL, argument);
}
int adoc_sem_wait(int * id){

  struct sembuf operations[1];
   /* Set up the sembuf structure. */
   /* Which semaphore in the semaphore array : */
    operations[0].sem_num = 0;
    /* Which operation? Subtract 1 from semaphore value : */
    operations[0].sem_op = -1;
    /* Set the flag so we will wait : */   
    operations[0].sem_flg = 0;

    /* So do the operation! */
    return semop(*id, operations, 1);



}



int adoc_sem_post(int * id){
struct sembuf operations[1];



/* Set up the sembuf structure. */
   /* Which semaphore in the semaphore array : */
    operations[0].sem_num = 0;
    /* Which operation? Add 1 to semaphore value : */
    operations[0].sem_op = 1;
    /* Set the flag so we will wait : */   
    operations[0].sem_flg = 0;

    /* So do the operation! */
    return semop(*id, operations, 1);
}

int adoc_sem_getvalue(int * id, int * sval){
#ifdef SEMUN_NOT_OK
    union semun {
		int val;
		struct semid_ds *buf;
		ushort * array;
	} argument;
#else
    union semun argument;
#endif
    
   *sval=semctl(*id,0,GETVAL,argument);
   return *sval;
}

int adoc_sem_destroy(int * id){
  #ifdef SEMUN_NOT_OK
    union semun {
		int val;
		struct semid_ds *buf;
		ushort * array;
	} argument;
#else
    union semun argument;
#endif
  return semctl(*id, 0, IPC_RMID, argument);
}



#else
int  adoc_sem_init(adoc_sem  *sem,  int  pshared,   unsigned   int val){
    /*manage only pshared==0*/
    if(pshared)
        return -1;
    /* build mutex for this sem*/
    pthread_mutex_init(&(sem->mutex),NULL);
    /*Init cond var*/
    pthread_cond_init(&(sem->cond),NULL);
    /*set the value*/
    sem->val=val;
    return 0;
}

int adoc_sem_getvalue(adoc_sem * sem, int * sval){
    pthread_mutex_lock(&(sem->mutex));
    *sval=sem->val;
    pthread_mutex_unlock(&(sem->mutex));
    return *sval;
}

int adoc_sem_post(adoc_sem * sem){
#ifdef DEBUG
    int r=rand();
    fprintf(stderr,"POST IN(%d) val=%d\n",r,sem->val);
#endif
    pthread_mutex_lock(&(sem->mutex));
    sem->val++;
    pthread_cond_broadcast(&(sem->cond));
    pthread_mutex_unlock(&(sem->mutex));
#ifdef DEBUG
    fprintf(stderr,"POST OUT(%d)\n",r);
#endif
    return 0;
}

int adoc_sem_wait(adoc_sem * sem){
#ifdef DEBUG
    int r=rand();
    fprintf(stderr,"WAIT IN (%d) val=%d\n",r,sem->val);
#endif
    pthread_mutex_lock(&(sem->mutex));
    while(sem->val<=0)
        pthread_cond_wait(&(sem->cond),&(sem->mutex));
    sem->val--;
    pthread_mutex_unlock(&(sem->mutex));
#ifdef DEBUG
    fprintf(stderr,"WAIT OUT(%d)\n",r);
#endif
    return 0;
}

int adoc_sem_destroy(adoc_sem *sem){
    pthread_mutex_lock(&(sem->mutex));
    pthread_cond_destroy(&(sem->cond));
    pthread_mutex_unlock(&(sem->mutex));
    pthread_mutex_destroy(&(sem->mutex));
    return 0;
}    

#endif





