#include <string.h>
#include "adoc.h"
#include "timings.h"
#include "config.h"
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <stdio.h>
#include <netinet/in.h>
#ifndef NO_SOCKET_TRICK
#include <netinet/ip.h>
#endif
#include <netinet/tcp.h>
#include <string.h>
#include <netdb.h>
#ifdef NO_SOCKET_TRICK
#include <strings.h>
#endif
#include <unistd.h>
#include <inttypes.h>
#include <netinet/tcp.h>
#include "client-server.h"

#undef LATENCY_ONLY

typedef int8_t byte;
typedef uint8_t ubyte;

int connect_server(int port,char *host){
  int sock;
  struct sockaddr_in nom;
  struct hostent *hp, *gethostbyname();
#ifndef NO_SOCKET_TRICK
  int tos;
#endif
  if((sock=socket(AF_INET,SOCK_STREAM,0))==-1){
    perror("Creating socket failed.\n");
    return -2;
  }
#ifndef NO_SOCKET_TRICK
  tos = IPTOS_THROUGHPUT;
  if(setsockopt(sock, IPPROTO_IP, IP_TOS, (char *)&tos,sizeof(tos))<0)
     printf("TOS not supported!\n");;
#endif


/*recherche de l'adresse internet du serveur*/


  if((hp=gethostbyname(host))==NULL){
    fprintf(stderr,"Host : %s unknown.\n",host);
    return -3;
  }
  

  bcopy(hp->h_addr,
	&nom.sin_addr,
	hp->h_length);
  nom.sin_family=AF_INET;
  nom.sin_port=htons(port);
  

  /* Asking for connection */

  if((connect(sock,(struct sockaddr *)&nom,sizeof(nom)))==-1){
    perror("Cannot connect to serveur 1\n");
    return -4;
  }

  return sock;
}
int main(int argc,char *argv[]){
    int port,socket;
    char *host;
    double time,time1,time2,time3;
    size_t sent_size,data_size;
    ubyte *buffer;
    ubyte tab[100];
    char msg[2];
    unsigned long int i,j,cc;
    int err_flag=0;
    CLOCK_T begin,end;
    if(argc!=3){
	fprintf(stderr,"usage : %s host port\n",argv[0]);
	exit(-1);
    }

  
    buffer=(ubyte *)malloc(BUFFER_SIZE*sizeof(ubyte));
    
    host=strdup(argv[1]);
    port=atoi(argv[2]);

    socket=connect_server(port,host);
    free(host);
    if(socket<0){
	fprintf(stderr,"Connection to server failed... Aborting.\n");
	exit(-1);
    }
#ifndef LATENCY_ONLY

        
    for(i=0;i<BUFFER_SIZE;i++)
        buffer[i]=i%10;
    
    printf("Mesuring network speed with disabled compression.\n");
    CLOCK(begin);
    cc=adoc_write_levels(socket,buffer,BUFFER_SIZE,&sent_size,0,0);
    CLOCK(end);
    time=CLOCK_DIFF(end,begin);
    if(cc!=BUFFER_SIZE){
	fprintf(stderr,"Error less byte writen (%ld) than it should (%ld)\n",(long)cc,(long)BUFFER_SIZE);
	free(buffer);
	exit(-1);
    }
    printf("Sending time = %.2f seconds.\n",time);
    printf("Data size= %lu.\n",(unsigned long int)cc);
    printf("Sent size = %lu.\n",(unsigned long int)sent_size);
    printf("Network throughput = %.2f Mbit/s.\n",sent_size*8/(time*1000000));
    printf("Application throughput = %.2f Mbit/s.\n",(unsigned long int)cc*8/(time*1000000));
    printf("Average compression ratio = %.2f.\n",(double)cc/(double)sent_size);
    CLOCK(begin);
    for(i=0;i<cc;i++)
	if(buffer[i]!=i%10){
	    fprintf(stderr,"buffer[%ld]=%d corrupted!\n",i,buffer[i]);
	    free(buffer);
	    return -1;
	}
    printf("No corruption!\n");
    
    printf("\nPress return to continue.");
    fgets(msg,2,stdin);

    
    
    printf("Mesuring network speed (normal mode).\n");
    CLOCK(begin);
    cc=adoc_write(socket,buffer,BUFFER_SIZE,&sent_size);
    CLOCK(end);  
    time=CLOCK_DIFF(end,begin);
    if(cc!=BUFFER_SIZE){
	fprintf(stderr,"Error less byte writen (%ld) than it should (%ld)\n",(long)cc,(long)BUFFER_SIZE);
	free(buffer);
	exit(-1);
    }
    printf("Sending time = %.2f seconds.\n",time);
    printf("Data size= %lu.\n",(unsigned long int)cc);
    printf("Sent size = %lu.\n",(unsigned long int)sent_size);
    printf("Network throughput = %.2f Mbit/s.\n",sent_size*8/(time*1000000));
    printf("Application throughput = %.2f Mbit/s.\n",(unsigned long int)cc*8/(time*1000000));
    printf("Average compression ratio = %.2f.\n",(double)cc/(double)sent_size);
    for(i=0;i<cc;i++)
	if(buffer[i]!=i%10){
	    fprintf(stderr,"buffer[%ld]=%d corrupted!\n",i,buffer[i]);
	    free(buffer);
	    return -1;
	}
    printf("No corruption!\n");
    
    printf("\nPress return to continue.");

    fgets(msg,2,stdin);
    printf("Mesuring network speed with forced compression.\n");
    

    CLOCK(begin);
    cc=adoc_write_levels(socket,buffer,BUFFER_SIZE,&sent_size,1,6);
    CLOCK(end);
    time=CLOCK_DIFF(end,begin);
    if(cc!=BUFFER_SIZE){
	fprintf(stderr,"Error less byte writen (%ld) than it should (%ld)\n",(long)cc,(long)BUFFER_SIZE);
	free(buffer);
	exit(-1);
    }
    printf("Sending time = %.2f seconds.\n",time);
    printf("Data size= %lu.\n",(unsigned long int)cc);
    printf("Sent size = %lu.\n",(unsigned long int)sent_size);
    printf("Network throughput = %.2f Mbit/s.\n",sent_size*8/(time*1000000));
    printf("Application throughput = %.2f Mbit/s.\n",(unsigned long int)cc*8/(time*1000000));
    printf("Average compression ratio = %.2f.\n",(double)cc/(double)sent_size);
    for(i=0;i<cc;i++)
	if(buffer[i]!=i%10){
	    fprintf(stderr,"buffer[%ld]=%d corrupted!\n",i,buffer[i]);
	    free(buffer);    
	    return -1;
	}
    printf("No corruption!\n");

    printf("Done!\n");
#endif
  printf("\nPress return to continue.");
   fgets(msg,2,stdin);

   printf("Timing latency (normal mode)\n");
   CLOCK(begin);
   for(i=0;i<50;i++){
       adoc_write(socket,buffer,1,&sent_size);
       printf(". ");fflush(stdout);
       adoc_read(socket,buffer,1);
   }

   CLOCK(end);
   time2=CLOCK_DIFF(end,begin);
   printf("\nMean duration of 1 byte ping pong : %.3f ms\n",time2*1000/50);

   printf("\nTiming latency with forced compression\n");
   CLOCK(begin);
   for(i=0;i<50;i++){
       adoc_write_levels(socket,buffer,1,&sent_size,1,6);
       printf(". ");fflush(stdout);
       adoc_read(socket,buffer,1);
   }
   CLOCK(end);
   time1=CLOCK_DIFF(end,begin);

   printf("\nMean duration of 1 byte ping pong : %.3f ms\n",time1*1000/50);

   printf("\nTiming latency with read/write system call\n");
   CLOCK(begin);
   for(i=0;i<50;i++){
       write(socket,buffer,1);
       printf(". ");fflush(stdout);
       read(socket,buffer,1);
   }
   CLOCK(end);
   time3=CLOCK_DIFF(end,begin);
   printf("\nMean duration of 1 byte ping pong  : %.3f ms\n",time3*1000/50);
#ifndef LATENCY_ONLY

    printf("Press return to continue.");
    fgets(msg,2,stdin);

    

    printf("\nChecking random ping-pong\n");
    
    srand(0);
    for(i=0;i<100;i++)
	tab[i]=(ubyte)rand(); 

    tab[0]=0;
    tab[1]=1;
    cc=adoc_write(socket,tab,100,&sent_size);

    for(i=0;i<100;i++){
	data_size=(size_t)tab[i];
	if(!(data_size%2)){
	    printf("s(%ld)",(long)data_size);fflush(stdout);
	    memset(buffer,1,data_size);
	    cc=adoc_write(socket,buffer,data_size,&sent_size);
	    printf(". ");fflush(stdout);
	    if(cc!=data_size){
		    err_flag++;
		fprintf(stderr,"\nError for step %lu (cc=%lu, shoud be %ld)\n",i,cc,(long)data_size);
	    }
	    
	    for(j=0;j<data_size;j++)
		if(buffer[j]!=1){
		    err_flag++;
		    fprintf(stderr,"\nbuffer[%lu]=%d corrupted!\n",j,buffer[j]);
		}
	}else{
	    printf("r(%ld)",(long)data_size);fflush(stdout);
	    bzero(buffer,data_size);
	    cc=adoc_read(socket,buffer,data_size);
	    printf(". ");fflush(stdout);
	    if(cc!=data_size){
		    err_flag++;
		fprintf(stderr,"\nError for step %lu (cc=%lu, shoud be %ld)\n",i,cc,(long)data_size);
	    }
	    
	    for(j=0;j<data_size;j++)
		if(buffer[j]!=1){
		    err_flag++;
		    fprintf(stderr,"\nbuffer[%lu]=%d error!\n",j,buffer[j]);
		}
	}
    }
   if(err_flag==0)
        printf("\nOK!\nNo error for ping-pong.\n");
   else{
	printf("\nNo ok! %d error(s)!\n",err_flag);
	free(buffer);    
	return -1;
   }

   printf("Done!\nPress return to continue.");
   fgets(msg,2,stdin);
printf("Checking internal buffer management without compression\n");
   for(i=0;i<100000;i++)
       buffer[i]=i%250;

   cc=adoc_write_levels(socket,buffer,100000,&sent_size,0,0);
   cc=adoc_write_levels(socket,buffer,100000,&sent_size,0,0);
   cc=adoc_write_levels(socket,buffer,100000,&sent_size,0,0);
   printf("Press return to continue.");
   fgets(msg,2,stdin);
   printf("Done!\nChecking internal buffer management with forced  compression\n");
   for(i=0;i<900000;i++)
       buffer[i]=i%250;

   cc=adoc_write_levels(socket,buffer,900000,&sent_size,1,9);
   cc=adoc_write_levels(socket,buffer,900000,&sent_size,1,9);
   cc=adoc_write_levels(socket,buffer,900000,&sent_size,1,9);
     
    printf("Done!\n");
    printf("Press return to continue.");
    fgets(msg,2,stdin);
    printf("Checking small buffer management\n");
    for(i=0;i<100000;i++)
       buffer[i]=i%250;

   cc=adoc_write(socket,buffer,100000,&sent_size);
    printf("Done!\n");
    printf("Press return to continue.");
    fgets(msg,2,stdin);
    printf("Checking small buffer management with forced compression\n");
   for(i=0;i<100000;i++)
       buffer[i]=i%250;

   cc=adoc_write_levels(socket,buffer,100000,&sent_size,1,6);
#endif
   adoc_close(socket);

   printf("Done!\nPlease check if all tests passed in the receiver terminal.\n\n");

   free(buffer);    
   return 0;

}
