#include "common.h"
#include "reception.h"

/* posting n w_sem after finishing processing n packets */
void post_w_sem(int n, ADOC_session *ADOC_global){
    int i;
    for(i=0;i<n;i++){
        /* Saying that one new packet can be added in the window*/
        adoc_sem_post(&(ADOC_global->w_sem));
    }
}


int read_recv_window(byte *type,byte *level,
                     size_t *data_size,byte** data_ptr,ADOC_session *ADOC_global){
    
    int a,b;
     adoc_sem_getvalue(&(ADOC_global->w_sem),&a);
     adoc_sem_getvalue(&(ADOC_global->r_sem),&b);

     /*printf("r_sem=%d\n",a);
     printf("w_sem=%d\n",b);*/
     /* wait for having at least one packet to read*/
    adoc_sem_wait(&(ADOC_global->r_sem));

    if(ADOC_global->error != 0) {
        return -1; 
    }

    (*type)=ADOC_global->header_window[ADOC_global->header_r].type;
    (*level)=ADOC_global->header_window[ADOC_global->header_r].level;
    (*data_size)=ADOC_global->header_window[ADOC_global->header_r].data_size;
    (*data_ptr)=ADOC_global->header_window[ADOC_global->header_r].data;

    /*Incrementing the reading pointer of the window cyclically*/
    ADOC_global->header_r++;
    if(ADOC_global->header_r>=ADOC_MAX_RECV_PACKETS)
        ADOC_global->header_r=0;

    return 0;
}


ssize_t zlib_uncompress_buffer(byte *uncomp_buffer,size_t uncomp_buffer_size,
                               byte *comp_buffer,ssize_t *comp_size, 
                               byte level,byte type,
                               size_t data_size,ADOC_session *ADOC_global){


    z_stream u_stream;
    byte val,*dummy_packet;
    int err2,err;


    u_stream.zalloc = (alloc_func)0;
    u_stream.zfree = (free_func)0;
    err = inflateInit(&u_stream);	
    CHECK_ERR(err, "inflateInit");    
    u_stream.next_out=(Bytef*)uncomp_buffer;
    u_stream.avail_out=(uInt)uncomp_buffer_size;

    u_stream.next_in=(Bytef*)comp_buffer;
    u_stream.avail_in=(uInt)data_size;
    err2=inflate(&u_stream,Z_SYNC_FLUSH);
    /* telling the reception thread that the packet is processed*/
    post_w_sem(1,ADOC_global);
    while(type!=ADOC_TYPE_LAST_PACKET){
        /* dummy_packet, is only for storing the adress of the next packet.
        Since compressed packet are always stored in a row, they formed
        a compressed buffer starting at adress comp_buffer, which is
        the only required adress for inflating data*/
        if(read_recv_window(&type,&val,&data_size,&dummy_packet,ADOC_global) < 0)
            return -1;
        (*comp_size)+=data_size;
        if((type!=ADOC_TYPE_LAST_PACKET)&&(type!=ADOC_TYPE_COMP)){
            fprintf(stderr,"AdOC : Protocol Error line %d!\nAborting...\n",__LINE__);
            ADOC_global->error=ADOC_PROTOCOL_ERROR;
            return -1;
        }
        if(level!=val){
            fprintf(stderr,"AdOC : Level Error : expected=%d, real=%d!\nAborting...\n",level,val);
            ADOC_global->error=ADOC_LEVEL_ERROR;
            return -1;
        }

        u_stream.avail_in=(uInt)data_size;
        err2=inflate(&u_stream,Z_SYNC_FLUSH);
        /* telling the reception thread that the packet is processed*/
        post_w_sem(1,ADOC_global);
    }
    if(err2!=Z_STREAM_END)
        CHECK_ERR(err2, "ADOC :Inflate Z_STREAM");
    inflateEnd(&u_stream);
    return uncomp_buffer_size-u_stream.avail_out;
}


ssize_t lzf_uncompress_buffer(byte *uncomp_buffer,size_t uncomp_buffer_size,
                              byte *comp_buffer,ssize_t *comp_size, 
                              byte level,byte type,
                              size_t data_size,ADOC_session *ADOC_global){


    byte val,*dummy_packet;
    ssize_t uncomp_size;
    int nb_packets=1;

    /* For lzf it is needed to retreive all the data before starting decompression*/
    while(type!=ADOC_TYPE_LAST_PACKET){
        /* dummy_packet, is only for storing the adress of the next packet.
        Since compressed packet are always stored in a row, they formed
        a compressed buffer starting at adress comp_buffer, which is
        the only required adress for inflating data*/
        if(read_recv_window(&type,&val,&data_size,&dummy_packet,ADOC_global) < 0)
            return -1;
        (*comp_size)+=data_size;
        if((type!=ADOC_TYPE_LAST_PACKET)&&(type!=ADOC_TYPE_COMP)){
            fprintf(stderr,"AdOC : Protocol Error line %d!\nAborting...\n",__LINE__);
            ADOC_global->error=ADOC_PROTOCOL_ERROR;
            return -1;
        }
        if(level!=val){
            fprintf(stderr,"AdOC : Level Error : expected=%d, real=%d!\nAborting...\n",level,val);
            ADOC_global->error=ADOC_LEVEL_ERROR;
            return -1;
        }
        nb_packets++;
    }

    uncomp_size=lzf_decompress (comp_buffer, *comp_size,uncomp_buffer,uncomp_buffer_size);
    if(uncomp_size==0){
        fprintf(stderr,"LZF : AdOC : Compression Error line %d (comp_size=%ld,uncomp_buffer_size=%ld) : %s!\nAborting...\n",__LINE__,(long)(*comp_size),(long)uncomp_buffer_size,strerror(errno));
        ADOC_global->error=ADOC_COMPRESSION_ERROR;
        return -1;
    }
    /*tell the receiving thread that it can use the n packets that stored the uncompressed data*/
    post_w_sem(nb_packets,ADOC_global);
    return uncomp_size;
}




ssize_t uncompress_buffer(byte *uncomp_buffer,size_t uncomp_buffer_size,
                          byte *comp_buffer,ssize_t *comp_size, 
                          byte level,byte type,
                          size_t data_size,ADOC_session *ADOC_global){


    if (level==1)
        return lzf_uncompress_buffer(uncomp_buffer,uncomp_buffer_size,comp_buffer,
                                     comp_size, level,type,data_size,ADOC_global);
    else{

        return zlib_uncompress_buffer(uncomp_buffer,uncomp_buffer_size,comp_buffer,
                                      comp_size, level,type,data_size,ADOC_global);

    }

}

void * file_decompression_thread(void *arg){

    byte level;
    size_t uncomp_buffer_size,size=0,recv_bytes=0,data_size;
    ssize_t res=0;
    ssize_t comp_size=0,uncomp_size=0;
    byte *uncomp_buffer,*comp_buffer,type,*dummy_packet;
    ssize_t *byte_recv;
    ADOC_session  *ADOC_global=(ADOC_session*)arg;
    FILE *pf;
    pf=ADOC_global->file;
    byte_recv=ADOC_global->byte_recv;



    if(read_recv_window(&type,&level,&uncomp_buffer_size,&dummy_packet,ADOC_global) < 0)
    {
        pthread_exit(0);
        return NULL;
    }
    post_w_sem(1,ADOC_global);
    if(type!=ADOC_TYPE_BS){
        fprintf(stderr,"ADOC Protocol Error line %d (type=%d)!\nAborting...\n",__LINE__,type);
        ADOC_global->error=ADOC_PROTOCOL_ERROR;
        pthread_exit(0);
    }
    uncomp_buffer=(byte *)malloc(uncomp_buffer_size);
    if(uncomp_buffer==NULL){
        errno=ENOMEM;
        ADOC_global->error=ADOC_ERROR_WITH_ERRNO_SET;
        pthread_exit(0);
    }
    do {
        if(read_recv_window(&type,&level,&data_size,&comp_buffer,ADOC_global) < 0)
            break;

        uncomp_size=comp_size=data_size;
        if(type==ADOC_TYPE_BS){
            post_w_sem(1,ADOC_global);
            fprintf(stderr,"ADOC Protocol Error line %d!\nAborting...\n",__LINE__);
            ADOC_global->error=ADOC_PROTOCOL_ERROR;
            free(uncomp_buffer);
            pthread_exit(0);
        }else if(type==ADOC_TYPE_EOF){
            post_w_sem(1,ADOC_global);
            break;
        }
        else if((type==ADOC_TYPE_COMP)||(type==ADOC_TYPE_LAST_PACKET)){
            if((uncomp_size=uncompress_buffer(uncomp_buffer,uncomp_buffer_size,
                                              comp_buffer,&comp_size,level,type,data_size,ADOC_global)) <0){
                free(uncomp_buffer);
                pthread_exit(0);
            }
            /*printf("\nuncomp_size=%ld\n",uncomp_size);*/
            res=WRITE(fileno(pf),uncomp_buffer,uncomp_size);
            if(res!=uncomp_size){
                if (res==-1){
                    ADOC_global->error=ADOC_ERROR_WITH_ERRNO_SET;
                    free(uncomp_buffer);
                    pthread_exit(0); 
                } else{
                    ADOC_global->error=ADOC_WRITE_ERROR;
                    free(uncomp_buffer);
                    pthread_exit(0); 
                }
            }
        } else if(type==ADOC_TYPE_UNCOMP){
            post_w_sem(1,ADOC_global);
            res=WRITE(fileno(pf),comp_buffer,comp_size);
            if(res!=comp_size){
                if (res==-1){
                    ADOC_global->error=ADOC_ERROR_WITH_ERRNO_SET;
                    free(uncomp_buffer);
                    pthread_exit(0); 
                } else{
                    ADOC_global->error=ADOC_WRITE_ERROR;
                    free(uncomp_buffer);
                    pthread_exit(0); 
                }
            }
        }else {
            post_w_sem(1,ADOC_global);
            fprintf(stderr,"AdOC :  Protocol error : type %d unknown!\nAborting...\n",type);
            ADOC_global->error=ADOC_PROTOCOL_ERROR;
            free(uncomp_buffer);
            pthread_exit(0);
        }
        size+=uncomp_size;
        recv_bytes+=comp_size;

        /*Check whenever an error has been generated by 
            the reception thread*/
        if(ADOC_global->error!=0){
            free(uncomp_buffer);
            pthread_exit(0);
        }
    }while(type!=ADOC_TYPE_EOF);
    free(uncomp_buffer);
    *byte_recv+=size;
    pthread_exit(0);
    /*Make compiler happy*/
    return NULL; 
}


void * data_decompression_thread(void *arg){

    byte level;
    size_t uncomp_buffer_size,size=0,recv_bytes=0,data_size;
    ssize_t comp_size=0,uncomp_size=0;
    byte *uncomp_buffer,*comp_buffer,type,*dummy_packet;
    size_t session_size,count,diff;
    byte *temp_buffer=NULL,*local_buffer=NULL,*start_local_buffer=NULL;
    ssize_t *byte_recv;
    int used_temp_buffer=0;
    ADOC_session  *ADOC_global=(ADOC_session*)arg;

    uncomp_buffer=ADOC_global->data_buffer;
    byte_recv=ADOC_global->byte_recv;
    count=ADOC_global->count;


    if(read_recv_window(&type,&level,&uncomp_buffer_size,&dummy_packet,ADOC_global) < 0)
    {
        pthread_exit(0);
        return NULL;
    }

    post_w_sem(1,ADOC_global);
    if(type!=ADOC_TYPE_BS){
        fprintf(stderr,"AdOC : Protocol Error line %d (type=%d)!\nAborting...\n",__LINE__,type);
        ADOC_global->error=ADOC_PROTOCOL_ERROR;
        pthread_exit(0);
    }

    if(read_recv_window(&type,&level,&session_size,&dummy_packet,ADOC_global) < 0)
    {
        pthread_exit(0);
        return NULL;
    }
    post_w_sem(1,ADOC_global);
    if(type!=ADOC_TYPE_SESSION_SIZE){
        fprintf(stderr,"AdOC : Protocol Error line %d!\nAborting...\n",__LINE__);
        ADOC_global->error=ADOC_PROTOCOL_ERROR;
        pthread_exit(0);
    }


    if(count<session_size){
        temp_buffer=(byte *)malloc(uncomp_buffer_size*sizeof(byte));
        if(temp_buffer==NULL){
            errno=ENOMEM;
            ADOC_global->error=ADOC_ERROR_WITH_ERRNO_SET;
            pthread_exit(0);
        }
        start_local_buffer=local_buffer=(byte *)malloc(session_size-count*sizeof(byte));
        if(start_local_buffer==NULL){
            errno=ENOMEM;
            ADOC_global->error=ADOC_ERROR_WITH_ERRNO_SET;
            free(temp_buffer);
            pthread_exit(0);
        }
    }

    do {
        if(read_recv_window(&type,&level,&data_size,&comp_buffer,ADOC_global) < 0)
            break;
        uncomp_size=comp_size=data_size;
        if((type==ADOC_TYPE_BS)||(type==ADOC_TYPE_SESSION_SIZE)){
            post_w_sem(1,ADOC_global);
            fprintf(stderr,"Adoc : Protocol Error line %d!\nAborting...\n",__LINE__);
            ADOC_global->error=ADOC_PROTOCOL_ERROR;
            if(count<session_size){
                free(temp_buffer);
                free(start_local_buffer);
                free(local_buffer);
            }
            pthread_exit(0);
        }else if(type==ADOC_TYPE_EOF){
            /*printf("EOF\n");*/
            break;
        }
        else if((type==ADOC_TYPE_COMP)||(type==ADOC_TYPE_LAST_PACKET)){
            if((count>=session_size)||(count>size+uncomp_buffer_size)){
                if((uncomp_size=uncompress_buffer(uncomp_buffer,uncomp_buffer_size,
                                                  comp_buffer,&comp_size,level,type,
                                                  data_size,ADOC_global))<0){
                    free(uncomp_buffer);
                    pthread_exit(0);
                }


                uncomp_buffer+=uncomp_size;
            }else if(!used_temp_buffer){
                /* The uncomp_buffer might be too small for receiving the entire session data*/
                /* We uncompress the received buffer into a temp one*/
                if((uncomp_size=uncompress_buffer(temp_buffer,uncomp_buffer_size,
                                                  comp_buffer,&comp_size,level,type,
                                                  data_size,ADOC_global)) < 0){
                    free(uncomp_buffer);
                    pthread_exit(0);
                }


                if(count-size<uncomp_size){/* if the uncomp buffer is effectivelly too small*/  
                    /* We split the temp buffer into two parts */
                    diff=count-size;
                    /* One goes into the uncomp buffer in order to fill it */ 
                    memcpy(uncomp_buffer,temp_buffer,diff);
                    uncomp_buffer+=diff;

                    /* The other part goes into the local buffer for further use */
                    memcpy(local_buffer,temp_buffer+diff,uncomp_size-diff);
                    local_buffer+=(uncomp_size-diff);
                    used_temp_buffer=1;
                }else{ /* there is enough space for storing the temp_buffer into the uncomp buffer*/
                    memcpy(uncomp_buffer,temp_buffer,uncomp_size);
                    uncomp_buffer+=uncomp_size;
                }

            } else {
                /*we are now decompressing all the data into the local buffer */ 
                if((uncomp_size=uncompress_buffer(local_buffer,uncomp_buffer_size,
                                                  comp_buffer,&comp_size,level,type,
                                                  data_size,ADOC_global)) <0){
                    free(uncomp_buffer);
                    pthread_exit(0);
                }

                local_buffer+=uncomp_size;

            }
        } else if(type==ADOC_TYPE_UNCOMP) {
            post_w_sem(1,ADOC_global);
            if((count>=session_size)||(count>size+comp_size)){
                memcpy(uncomp_buffer,comp_buffer,comp_size);
                uncomp_buffer+=comp_size;
            }else if(!used_temp_buffer){
                diff=count-size;
                /* We split the comp_buffer into two parts */
                /* One goes into the uncomp buffer in order to fill it */ 
                memcpy(uncomp_buffer,comp_buffer,diff);
                /* The other part goes into the local buffer for further use */
                memcpy(local_buffer,comp_buffer+diff,comp_size-diff);
                local_buffer+=(comp_size-diff);
                used_temp_buffer=1;	
            }else{
                /*we are now copying all the data into the local buffer */ 
                memcpy(local_buffer,comp_buffer,comp_size);
                /*printf("\nuncomp_size=%ld\n",comp_size);*/
                local_buffer+=comp_size;		
            }
        } else {
            post_w_sem(1,ADOC_global);
            fprintf(stderr,"AdOC : Protocol error : type %d unknown!\nAborting...\n",type);
            ADOC_global->error=ADOC_PROTOCOL_ERROR;
            if(count<session_size){
                free(temp_buffer);
                free(start_local_buffer);
                free(local_buffer);
            }
            pthread_exit(0);
        }
        size+=uncomp_size;
        recv_bytes+=comp_size;
        /*Check whenever an error has been generated by 
            the reception thread*/
        if(ADOC_global->error!=0){
            if(count<session_size){
                free(temp_buffer);
                free(start_local_buffer);
                free(local_buffer);
            }
            pthread_exit(0);
        }

    }while(type!=ADOC_TYPE_EOF);

    if(session_size!=size){
        fprintf(stderr,"AdOC : network error (received = %lu, expected = %lu)!\n",(long)size,(long)session_size);
        ADOC_global->error=ADOC_NETWORK_ERROR;
        if(count<session_size){
            free(temp_buffer);
            free(start_local_buffer);
            free(local_buffer);
        }
        pthread_exit(0);
    }

    *byte_recv+=ADOC_MIN(count,session_size);

    if(count<session_size){
        free(temp_buffer);
        store_buffer(ADOC_global->socket,start_local_buffer,session_size-count,ADOC_global);

        /* Do not free the start_local_buffer and the local buffer because they are
            saved for further retreive
            Therefore it is a bug to uncomment the two following lines...
            free(start_local_buffer);
        free(local_buffer);*/
    }

    pthread_exit(0);
    /*Make compiler happy*/
    return NULL; 
}




