#include "common.h"
#include "emission.h"
#include "timings.h"

int adoc_em_init(ADOC_session *ADOC_global){
    int rval,i;
    /*shared objects initialization*/
    rval=adoc_sem_init (&(ADOC_global->history_sem), 0, 0);
    CHECK_RETVAL(rval,"sem_init");
    rval=adoc_sem_init (&(ADOC_global->r_sem), 0, 0);
    CHECK_RETVAL(rval,"sem_init");
    

    rval=adoc_sem_init (&(ADOC_global->w_sem), 0, ADOC_MAX_NB_PACKETS-1);
    CHECK_RETVAL(rval,"sem_init");

    
    CLOCK(ADOC_global->time_stamp);


    ADOC_global->speed_min=(double*)malloc((ADOC_global->max_level+1)*sizeof(double));
    if(ADOC_global->speed_min==NULL){
	errno=ENOMEM;
	ADOC_global->error=ADOC_ERROR_WITH_ERRNO_SET;
	free(ADOC_global->header_window);
	return -1;
    }

    for(i=0;i<ADOC_global->max_level+1;i++){
        ADOC_global->speed_min[i]=-1;
    }

    
    
    
    ADOC_global->reset_speed_min=(int*)malloc((ADOC_global->max_level+1)*sizeof(int));
    if(ADOC_global->reset_speed_min==NULL){
	errno=ENOMEM;
	ADOC_global->error=ADOC_ERROR_WITH_ERRNO_SET;
	free(ADOC_global->header_window);
	return -1;
    }
    for(i=0;i<ADOC_global->max_level+1;i++)
        ADOC_global->reset_speed_min[i]=1;




    
    pthread_mutex_init(&(ADOC_global->fifo_mutex),0);
    /* With LZF we assume that the maximum expansion is not greater than twice the original size
     Therefore the fifo can not be completely full because we need some space to compress more 
     than one packet in a row (at most(2*LZF_BUFFER_SIZE)/ADOC_PACKET_SIZE). */
    ADOC_global->header_window=(header*)malloc(sizeof(header)*(ADOC_MAX_SEND_PACKETS+1+(2*LZF_BUFFER_SIZE)/ADOC_PACKET_SIZE));
    if(ADOC_global->header_window==NULL){
	errno=ENOMEM;
	ADOC_global->error=ADOC_ERROR_WITH_ERRNO_SET;
	return -1;
    }

    /* the data window must be large eanough to store at least ADOC_MAX_SEND_PACKETS plus
       the largest possible compressed buffer which is at most twice the size of an 
       uncompressed buffer. This ensures that after ADOC_data_w_pointer there is
       always enough space to store a compressed buffer. Indeed, ADOC_data_w_pointer is
       incremented by a packet size until the algorithm detects that its address is
       greater than ADOC_data_window+ADOC_PACKET_SIZE*ADOC_MAX_SEND_PACKETS. The detetction
       test is perfromed after sending an uncompressed packet, a control packet or
       the last packet of a compressed buffer. data_r_window is re-initialized to 
       ADOC_data_window if there is a possibility of lacking of memory space.
    */
    ADOC_global->data_window=(byte *)malloc(ADOC_PACKET_SIZE*(ADOC_MAX_SEND_PACKETS)+2*ADOC_BUFFER_SIZE);
    if(ADOC_global->data_window==NULL){
	errno=ENOMEM;
	ADOC_global->error=ADOC_ERROR_WITH_ERRNO_SET;
	free(ADOC_global->header_window);
	return -1;
    }

    ADOC_global->header_r=0;
    ADOC_global->header_w=0;
    ADOC_global->data_w_pointer=ADOC_global->data_window;
    ADOC_global->fifo_size=0;
    ADOC_global->error=0; /*No error so far...*/
    return 0;
}

void adoc_send_finalyze(ADOC_session *ADOC_global){
    int rval;
    free(ADOC_global->header_window);
   free(ADOC_global->data_window);
   free(ADOC_global->speed_min);
   free(ADOC_global->reset_speed_min);
   rval=adoc_sem_destroy (&(ADOC_global->history_sem));
    CHECK_RETVAL(rval,"sem_destroy");
    rval=adoc_sem_destroy(&(ADOC_global->r_sem));
    CHECK_RETVAL(rval,"sem_destroy");
    rval=adoc_sem_destroy (&(ADOC_global->w_sem));
    CHECK_RETVAL(rval,"sem_destroy");
    pthread_mutex_destroy(&(ADOC_global->fifo_mutex));
    
}





ssize_t time_raw_send_file_in_pieces(int socket, FILE *pf,
                                     size_t *sent_size,
                                     double *throughput){

    uint32_t sval;
    byte c;
    int i;
    byte buffer[TCP_PIECE_SIZE];
    size_t count,data_size=TCP_TIMING_SIZE;
    double speed,speed_max=0,time;
    CLOCK_T begin,end;
    size_t prev_sent_size;

    c=ADOC_RAW;
    WRITE(socket,&c,1);
    sval=htonl(data_size);
    WRITE(socket,&sval,4);

    i=0;
    prev_sent_size=0;
    CLOCK(begin);
    do{


        count=fread(buffer,1,TCP_PIECE_SIZE,pf);
        *sent_size+=WRITE(socket,buffer,count);
        i++;
        if((i%8)==0){
            CLOCK(end);
            time=CLOCK_DIFF(end,begin);
            /* evaluate the speed of the network*/
            speed=((double)((*sent_size)-prev_sent_size)*8)/(time*1000000); /*in Mbit/s*/
            if(speed>speed_max)
                speed_max=speed;
            prev_sent_size=*sent_size;
            CLOCK(begin);
        }
    }while(*sent_size<data_size);
    CLOCK(end);
    time=CLOCK_DIFF(end,begin);
    /* evaluate the speed of the network*/
    speed=((double)((*sent_size)-prev_sent_size)*8)/(time*1000000); /*in Mbit/s*/
    if(speed>speed_max)
        speed_max=speed;

    *throughput=speed_max;

    return *sent_size;

}


ssize_t raw_send_file(int socket, FILE *pf,
                      size_t data_size,
                      size_t *sent_size){
    uint32_t sval;
    size_t count;
    byte c;
    byte buffer[ADOC_PACKET_SIZE];
    if(data_size<127){
        c=(byte)data_size;
        WRITE(socket,(byte *)&c,1);
    }else{
        c=ADOC_RAW;
        WRITE(socket,(byte *)&c,1);
        sval=htonl(data_size);
        WRITE(socket,(byte *)&sval,4);
    }
    while((count=fread(buffer,1,ADOC_PACKET_SIZE,pf))!=0){
        *sent_size+=WRITE(socket,buffer,count);
        data_size-=count;
        if(data_size<=0)
            return *sent_size;
    }
    return *sent_size;

}

ssize_t  adoc_send_file_levels(int socket, FILE *pf,
		    size_t *sent_size,unsigned int min_level,unsigned int max_level){
    
    
    pthread_t th1,th2;
    void *ret;
    ADOC_session ADOC_global;
    struct stat inode;
    byte c;
    double speed;
    int res=0;
    size_t remaining_size;
                
    /*size of the file*/    
    fstat(fileno(pf),&inode);
    ADOC_global.file_size=inode.st_size;
    *sent_size=ADOC_global.sent_size=0;


    /*checking compression level*/
    if(min_level<ADOC_MIN_LEVEL)
        min_level=ADOC_MIN_LEVEL;
    if(max_level>ADOC_MAX_LEVEL)
        max_level=ADOC_MAX_LEVEL;
    if(min_level>max_level)
        min_level=max_level;
    
    
    /*Can we do raw send? */
    if (((ADOC_global.file_size<2*TCP_TIMING_SIZE)&&(min_level==0))||(max_level<1)){
    return raw_send_file(socket,pf,ADOC_global.file_size,sent_size);
    }else{
        if (min_level>0){ /* if forced compression */
            c=ADOC_CMP;
            WRITE(socket,(byte *)&c,1);
        }else{
            c=ADOC_TEST;
            WRITE(socket,(byte *)&c,1);

            /* time a short connection*/
            res=time_raw_send_file_in_pieces(socket,pf,sent_size,&speed);

            remaining_size=ADOC_global.file_size-res;
            if((speed>CMP_THRESHOLD)||(remaining_size<TCP_TIMING_SIZE)){
                return raw_send_file(socket,pf,remaining_size,sent_size);
            }else{
                c=ADOC_CMP;
                WRITE(socket,(byte *)&c,1);
                ADOC_global.sent_size+=res;
                ADOC_global.file_size-=res;
            }
        }
    }

    



   ADOC_global.file=pf;
   ADOC_global.socket=socket;
   ADOC_global.min_level=min_level;
   ADOC_global.max_level=max_level;

   if(adoc_em_init(&ADOC_global)!=0)
	return ADOC_global.error;
   
    if (pthread_create (&th1, NULL, file_compression_thread, &ADOC_global) < 0) {
	 fprintf(stderr, "pthread_create error for compression thread\n");
	return ADOC_ERROR_WITH_ERRNO_SET;
    }
    
    if (pthread_create (&th2, NULL, emission_thread,&ADOC_global) < 0) {
	fprintf (stderr, "pthread_create error for compression thread\n");
	return ADOC_ERROR_WITH_ERRNO_SET;
    }

    (void)pthread_join (th1, &ret);
    (void)pthread_join (th2, &ret);

    adoc_send_finalyze(&ADOC_global);
    *sent_size=ADOC_global.sent_size;
    if(ADOC_global.error==0)
	return ADOC_global.file_size+res;
    else 
	return ADOC_global.error;
}


ssize_t launch_em_threads(int socket, byte *data_buffer,
                      size_t data_size,
                      size_t *sent_size,unsigned int min_level,unsigned int max_level){
    pthread_t th1,th2;
    void *ret;
    ADOC_session ADOC_global;

    ADOC_global.sent_size=*sent_size;

    ADOC_global.data_buffer=data_buffer;
    ADOC_global.data_size=data_size;
    ADOC_global.socket=socket;
    ADOC_global.min_level=min_level;
    ADOC_global.max_level=max_level;
    
    if(adoc_em_init(&ADOC_global)!=0)
        return ADOC_global.error;

    if (pthread_create (&th1, NULL, data_compression_thread, &ADOC_global) < 0) {
        fprintf (stderr, "pthread_create error for compression thread\n");
        return ADOC_ERROR_WITH_ERRNO_SET;
    }

    if (pthread_create (&th2, NULL, emission_thread,&ADOC_global) < 0) {
        fprintf (stderr, "pthread_create error for emission thread\n");
        return ADOC_ERROR_WITH_ERRNO_SET;
    }

    (void)pthread_join (th1, &ret);
    (void)pthread_join (th2, &ret);

    adoc_send_finalyze(&ADOC_global);

    *sent_size=ADOC_global.sent_size;
    if(ADOC_global.error==0)
        return ADOC_global.uncomp_size;
    else
        return ADOC_global.error;
}

    


ssize_t adoc_write(int socket, void *data_buffer,
                   size_t data_size,
                   size_t *sent_size){
    ssize_t res;
    res=adoc_write_levels(socket,data_buffer,data_size,sent_size,ADOC_MIN_LEVEL,ADOC_MAX_LEVEL);


    /*debug timing 
      {
    CLOCK_T date;

    CLOCK(date);
    CLOCK_DISPLAY(date); fprintf(stderr," : writen : %ld\n",(long int)res);
    usleep(10000);
    }
    */
    return res;
}

ssize_t  adoc_send_file(int socket, FILE *pf,
                        size_t *sent_size){

    return adoc_send_file_levels(socket,pf,sent_size,ADOC_MIN_LEVEL,ADOC_MAX_LEVEL);
}



