#include "common.h"
#include "emission.h"
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/tcp.h>

ssize_t raw_send(int socket, byte *data_buffer,
                 size_t data_size,
                 size_t *sent_size){

    uint32_t sval;
    int8_t c;
    if(data_size<127){
        c=(int8_t)data_size;
        WRITE(socket,&c,1);
        *sent_size+=WRITE(socket,data_buffer,data_size);
    }else{
        c=ADOC_RAW;
        WRITE(socket,&c,1);
        sval=htonl(data_size);
        WRITE(socket,&sval,4);
        *sent_size+=WRITE(socket,data_buffer,data_size);
    }
    return *sent_size;

}


ssize_t time_raw_send_in_pieces(int socket, byte *data_buffer,
                           size_t *sent_size,
                           double *throughput){

    uint32_t sval;
    size_t data_size=TCP_TIMING_SIZE;
    int i;
    double speed,speed_max=0,time;
    CLOCK_T begin,end;
    size_t prev_sent_size,count;
        

     sval=htonl(data_size);
     WRITE(socket,&sval,4);

     i=0;
     prev_sent_size=0;
     CLOCK(begin);
     do{
         count=ADOC_MIN(data_size-(*sent_size),TCP_PIECE_SIZE);
         (*sent_size)+=WRITE(socket,data_buffer+(*sent_size),count);

         i++;
         if((i%8)==0){
             CLOCK(end);
             time=CLOCK_DIFF(end,begin);
             /* evaluate the speed of the network*/
             speed=((double)((*sent_size)-prev_sent_size)*8)/(time*1000000); /*in Mbit/s*/
             /*printf("time=%.3f,speed=%.3f\n",time,speed);*/
             if(speed>speed_max)
                 speed_max=speed;
             prev_sent_size=*sent_size;
             CLOCK(begin);
         }
     }while(*sent_size<data_size);
     CLOCK(end);
     time=CLOCK_DIFF(end,begin);
     /* evaluate the speed of the network*/
     if((*sent_size)-prev_sent_size){
         speed=((double)((*sent_size)-prev_sent_size)*8)/(time*1000000); /*in Mbit/s*/
         /*printf("time=%.3f,speed=%.3f\n",time,speed);*/
         if(speed>speed_max)
             speed_max=speed;
     }

     *throughput=speed_max;

    return *sent_size;

}






ssize_t adoc_write_levels_2(int socket, void *buf,
                          size_t data_size,
                          size_t *sent_size,unsigned int min_level,unsigned int max_level){


    byte c;
    double speed;
    int res=0;
    byte *data_buffer;
    *sent_size=0;

    data_buffer=(byte*)buf;

    /*checking compression level*/
    if(min_level<ADOC_MIN_LEVEL)
        min_level=ADOC_MIN_LEVEL;
    if(max_level>ADOC_MAX_LEVEL)
        max_level=ADOC_MAX_LEVEL;
    if(min_level>max_level)
        min_level=max_level;
    
    



    /*Can we do raw send? */
    if (((data_size<2*TCP_TIMING_SIZE)&&(min_level==0))||(max_level<1)){
        ssize_t res;
        res=raw_send(socket,data_buffer,data_size,sent_size);
        return res;
    }else{
        if (min_level>0){ /* if forced compression */
            c=ADOC_CMP;
            WRITE(socket,&c,1);
        }else{
            c=ADOC_TEST;
            WRITE(socket,&c,1);
            /* time a short connection*/
            res=time_raw_send_in_pieces(socket,data_buffer,sent_size,&speed);
            /*printf("res=%d\n",res);*/
            data_size-=res;
            data_buffer+=res;

            /*printf("Speed=%f(theshold : %d)\n",speed,CMP_THRESHOLD );*/
            if((speed>CMP_THRESHOLD)||(data_size<TCP_TIMING_SIZE)){
                return raw_send(socket,data_buffer,data_size,sent_size);
            }else{
                c=ADOC_CMP;
                WRITE(socket,&c,1);

            }
        }
    }
return res+launch_em_threads(socket,data_buffer,data_size,sent_size, min_level, max_level);
}


/*Test if the Nagle algorithm is on. 
  Return O if no*/
int nagle_algorithm(int sock){
  socklen_t opt_size;
  int on;

  opt_size=sizeof(int);
  getsockopt(sock,IPPROTO_TCP,TCP_NODELAY,&on,&opt_size);

  return on;
}


/*disable the Nagle algorithm*/
void disable_nagle_algorithm(int sock){
  const int on =1;
  setsockopt(sock,IPPROTO_TCP,TCP_NODELAY,&on,sizeof(on));
  

}

/*enable the Nagle algorithm*/
void enable_nagle_algorithm(int sock){
  const int on =0;
  setsockopt(sock,IPPROTO_TCP,TCP_NODELAY,&on,sizeof(on));
}


ssize_t adoc_write_levels(int socket, void *buf,
                          size_t data_size,
                          size_t *sent_size,unsigned int min_level,unsigned int max_level){
  ssize_t res;
  size_t sent_size_dummy;
  
  if(sent_size==NULL)
    sent_size=&sent_size_dummy;
  


  /*disable the Nagle algorithm for small size data if necessary*/
  if((!nagle_algorithm(socket))&&(data_size<8192)){
    disable_nagle_algorithm(socket);
    res=adoc_write_levels_2(socket,buf,data_size,sent_size,min_level,max_level);
    enable_nagle_algorithm(socket);
    return res;
  }else{
    return adoc_write_levels_2(socket,buf,data_size,sent_size,min_level,max_level);
  }
}
    

